# SPDX-FileCopyrightText: ☭ Emery Hemingway
# SPDX-License-Identifier: Unlicense

import
  std/[httpcore, options, parseutils, sets, streams, strutils, tables, times, uri],
  pkg/taps,
  pkg/preserves,
  ../../syndicate, ../bags, ./timers, ../protocols/http

const
  CRLF = "\x0d\x0a"
  SP = { ' ', '\x09', '\x0b', '\x0c', '\x0d' }
  SupportedVersion = "HTTP/1.1"
  IMF = initTimeFormat"ddd, dd MMM yyyy HH:mm:ss"

when defined(posix):
  proc echo(args: varargs[string, `$`]) {.used.} =
    stderr.writeLine(args)

type
  Driver = ref object
    facet: Facet
    ds, timers: Cap
    sequenceNumber: BiggestInt
  Session = ref object
    facet: Facet
    driver: Driver
    conn: Connection
    exch: Exchange
    pendingLen: int
    port: Port
  Exchange = ref object of Entity
    cap: Cap
    ses: Session
      # TODO: store what we need fom Session locally.
    req: HttpRequest
    binding: Option[HttpBinding]
    stream: StringStream
    contentType: string
    contentLen: int
    mode: HttpResponseKind
    active: bool

proc badRequest(conn: Connection; msg: string) =
  conn.send(SupportedVersion & " " & msg, endOfMessage = true)

proc extractQuery(s: var string): Table[Symbol, seq[QueryValue]] =
  let start = succ skipUntil(s, '?')
  if start < s.len:
    var query = s[start..s.high]
    s.setLen(pred start)
    for key, val in uri.decodeQuery(query):
      var list = result.getOrDefault(Symbol key)
      list.add QueryValue(orKind: QueryValueKind.string, string: val)
      result[Symbol key] = list

proc parseRequest(conn: Connection; exch: Exchange; text: string): int =
  ## Parse an `HttpRequest` request out of a `text` from a `Connection`.
  var
    token: string
    off: int

  template advanceSp =
    let n = skipWhile(text, SP, off)
    if n < 1:
      badRequest(conn, "400 invalid request")
      return
    inc(off, n)

  # method
  off.inc parseUntil(text, token, SP, off)
  exch.req.method = token.toLowerAscii.Symbol
  advanceSp()

  # target
  if text[off] == '/': inc(off) #TODO: always a leading slash?
  off.inc parseUntil(text, token, SP, off)
  advanceSp()

  block:
    var version: string
    off.inc parseUntil(text, version, SP, off)
    advanceSp()
    if version != SupportedVersion:
      badRequest(conn, "400 version not supported")
      return

  exch.req.query = extractQuery(token)

  if token != "":
    exch.req.path = split(token, '/')
    for p in exch.req.path.mitems:
      # normalize the path
      for i, c in p:
        if c in {'A'..'Z'}:
          p[i] = char c.ord + 0x20

  exch.req.host = RequestHost(orKind: RequestHostKind.absent)

  template advanceLine =
    inc off, skipWhile(text, {'\x0d'}, off)
    if text.high < off or text[off] != '\x0a':
      badRequest(conn, "400 invalid request")
      return
    inc off, 1

  advanceLine()
  while off < text.len:
    off.inc parseUntil(text, token, {'\x0d', '\x0a'}, off)
    if token == "": break
    advanceLine()
    var
      (key, vals) = httpcore.parseHeader(token)
      k = key.toLowerAscii
      v = exch.req.headers.getOrDefault(cast[Symbol](k))
    for e in vals.mitems:
      var e = e.move.toLowerAscii
      case k
      of "host":
        exch.req.host = RequestHost(orKind: RequestHostKind.`present`, present: v)
      of "content-length":
        discard parseInt(e, exch.contentLen)
        if exch.contentLen > (1 shl 23):
          badRequest(conn, "413 Content Too Large")
        if exch.contentLen > 0:
          exch.req.body = Value(
              kind: pkByteString, bytes: newSeqOfCap[byte](exch.contentLen))
      of "content-type":
        exch.contentType = e
      if v == "":
        v = e.toLowerAscii
      else:
        v.add ", "
        v.add e.toLowerAscii
      exch.req.headers[cast[Symbol](k)] = v
  advanceLine()
  result = off

proc len(chunk: Chunk): int =
  case chunk.orKind
  of ChunkKind.string: chunk.string.len
  of ChunkKind.bytes: chunk.bytes.len

proc lenLine(chunk: Chunk): string =
  result = chunk.len.toHex.strip(true, false, {'0'})
  result.add CRLF

proc send[T: byte|char](ses: Session; data: openarray[T]) =
  ses.conn.send(addr data[0], data.len, endOfMessage = false)

proc send(ses: Session; chunk: Chunk) =
  case chunk.orKind
  of ChunkKind.string:
    ses.send(chunk.string)
  of ChunkKind.bytes:
    ses.send(chunk.bytes)

func isTrue(v: Value): bool = v.kind == pkBoolean and v.bool

proc dispatch(exch: Exchange; turn: Turn; res: HttpResponse) =
  case res.orKind
  of HttpResponseKind.status:
    if exch.mode == res.orKind:
      exch.active = true
      exch.ses.conn.startBatch()
      exch.stream.write(
          SupportedVersion, " ", res.status.code, " ", res.status.message, CRLF &
          # "connection: close" & CRLF &
            # RFC9112 says we SHOULD support persistent connections.
          "date: ", now().format(IMF), CRLF
            # add Date header automatically - RFC 9110 Section 6.6.1.
        )
      exch.mode = HttpResponseKind.header

  of HttpResponseKind.header:
    if exch.mode == res.orKind:
      exch.stream.write(res.header.name, ": ", res.header.value, CRLF)

  of HttpResponseKind.chunk:
    if res.chunk.chunk.len > 0:
      if exch.mode == HttpResponseKind.header:
        exch.stream.write("transfer-encoding: chunked" & CRLF & CRLF)
        exch.ses.send(move exch.stream.data)
        exch.mode = res.orKind
      if exch.mode == res.orKind:
        exch.ses.send(res.chunk.chunk.lenLine)
        exch.ses.send(res.chunk.chunk)
        exch.ses.send(CRLF)

  of HttpResponseKind.done:
    if exch.mode in {HttpResponseKind.header, HttpResponseKind.chunk}:
      if exch.mode == HttpResponseKind.header:
        exch.stream.write("content-length: ", $res.done.chunk.len & CRLF & CRLF)
        exch.ses.send(move exch.stream.data)
        if res.done.chunk.len > 0:
          exch.ses.send(res.done.chunk)
      elif exch.mode == HttpResponseKind.chunk:
        exch.ses.send(res.done.chunk.lenLine)
        if res.done.chunk.len > 0:
          exch.ses.send(res.done.chunk)
        exch.ses.send(CRLF & "0" & CRLF & CRLF)
      exch.mode = res.orKind
      exch.ses.conn.endBatch()
      if exch.req.headers.getOrDefault(Symbol"connection") == "close":
        exch.ses.conn.close()
      stop(turn)
        # stop the facet scoped to the exchange
        # so that the response capability is withdrawn

proc scheduleTimeout(exch: Exchange; turn: Turn) =
  const timeout = initDuration(seconds = 4)
  after(turn, exch.ses.driver.timers, timeout) do (turn: Turn):
    if not exch.active:
      var res = HttpResponse(orKind: HttpResponseKind.status)
      res.status.code = 504
      res.status.message = "Binding timeout"
      exch.dispatch(turn, res)
      res = HttpResponse(orKind: HttpResponseKind.done)
      exch.dispatch(turn, res)
        # Use a HttpResponse to reuse logic.

method message(exch: Exchange; turn: Turn; a: AssertionRef) =
  # Send responses back into a connection.
  if a.value.isTrue:
    # Sync message from dataspace.
    exch.scheduleTimeout(turn)
    if exch.binding.isSome:
      # The best binding in the dataspace has been selected.
      let handler = exch.binding.get.handler.unembed(Cap)
      if handler.isSome:
        # Send the handler the exchange capability.
        if exch.contentType.startsWith "application/json":
          if exch.req.body.isByteString:
            var bodyBytes = exch.req.body.bytes.move
            exch.req.body = cast[string](bodyBytes).parsePreserves
        discard publish(turn, handler.get, HttpContext(
            req: exch.req,
            res: exch.cap.embed,
          ))
      else:
        exch.binding.reset()
  else:
    var res: HttpResponse
    if exch.mode != HttpResponseKind.done and res.fromPreserves a.value:
      exch.dispatch(turn, res)

func `==`(s: string; rh: RequestHost): bool =
  rh.orKind == RequestHostKind.present and rh.present == s

proc match(b: HttpBinding, r: HttpRequest): bool =
  ## Check if `HttpBinding` `b` matches `HttpRequest` `r`.
  result =
    (b.host.orKind == HostPatternKind.any or
      b.host.host == r.host) and
    (b.port == r.port) and
    (b.method.orKind == MethodPatternKind.any or
        b.method.specific == r.method)
  if result:
    for i, p in b.path:
      if i > r.path.high: return false
      case p.orKind
      of PathPatternElementKind.wildcard: discard
      of PathPatternElementKind.label:
        if p.label != r.path[i]: return false
      of PathPatternElementKind.rest:
        return i == b.path.high
          # return false if "..." isn't the last element

proc strongerThan(a, b: HttpBinding): bool =
  ## Check if `a` is a stronger `HttpBinding` than `b`.
  result =
    (a.host.orKind != b.host.orKind and
      a.host.orKind == HostPatternKind.host) or
    (a.method.orKind != b.method.orKind and
      a.method.orKind == MethodPatternKind.specific)
  if not result:
    if a.path.len > b.path.len: return true
    for i in b.path.low..a.path.high:
      if a.path[i].orKind != b.path[i].orKind and
          a.path[i].orKind == PathPatternElementKind.label:
        return true

proc service(turn: Turn; exch: Exchange) =
  ## Service an HTTP message exchange.
  let pat = grab(HttpBinding?:{
      0: drop(), # match all hosts
      1: ?exch.req.port,
      2: drop(), # match all methods
      3: drop(), # match all paths
      4: drop(),
    })
  onPublish(turn, exch.ses.driver.ds, pat) do (b: HttpBinding):
    if b.match exch.req:
      if exch.binding.isNone or b.strongerThan exch.binding.get:
        exch.binding = some b
          # found a better binding
  exch.cap =  turn.newCap(exch)
  sync(turn, exch.ses.driver.ds, exch.cap)
    # When the sync returns we have seen all the bindings.

proc exchange(ses: Session) =
  ## Detach the current Exchange from the session
  ## and pass it into Syndicate.
  ses.pendingLen = 0
  var exch = move ses.exch
  exch.ses = ses
  ses.facet.run do (turn: Turn):
    inFacet(turn) do (turn: Turn):
      # start a new facet for this message exchange
      preventInertCheck(turn)
      exch.facet = turn.facet
      exch.stream = newStringStream()
      exch.mode = HttpResponseKind.status
      turn.service(exch)

proc service(ses: Session) =
  ## Service a connection to an HTTP client.
  const oneMiB = 1 shl 20
  ses.facet.onStop do (turn: Turn):
    close ses.conn
  ses.conn.onClosed do ():
    stop ses.facet
  ses.conn.onReceivedPartial do (data: seq[byte]; ctx: MessageContext; eom: bool):
    if ses.pendingLen == 0:
      if ses.exch.isNil:
        new ses.exch
      let off = parseRequest(ses.conn, ses.exch, cast[string](data))
      if off > 0:
        assert not ses.exch.isNil
        inc(ses.driver.sequenceNumber)
        ses.exch.req.sequenceNumber = ses.driver.sequenceNumber
        ses.exch.req.port = BiggestInt ses.port
        ses.pendingLen = ses.exch.contentLen
        if off < data.len:
          let n = min(data.len - off, ses.pendingLen)
          ses.exch.req.body.bytes.add data[off .. off+n.pred]
          ses.pendingLen.dec n
    else:
      let n = min(data.len, ses.pendingLen)
      ses.exch.req.body.bytes.add data[0..n.pred]
      ses.pendingLen.dec n
    assert ses.pendingLen >= 0, $ses.pendingLen
    if ses.pendingLen == 0:
      ses.exchange()
      ses.conn.receive(maxLength = oneMiB)
    else:
      ses.conn.receive(maxLength=ses.pendingLen)
  ses.conn.receive(maxLength =oneMiB)

proc newListener(port: Port): Listener =
  var lp = newLocalEndpoint()
  lp.with port
  listen newPreconnection(local=[lp])

proc httpListen(turn: Turn; driver: Driver; port: Port): Listener =
  let facet = turn.facet
  var listener = newListener(port)
  preventInertCheck(turn)
  listener.onListenError do (err: ref Exception):
    terminateFacet(facet, err)
  facet.onStop do (turn: Turn):
    stop listener
  echo "listening for HTTP on port ", port
  listener.onConnectionReceived do (conn: Connection):
    driver.facet.run do (turn: Turn):
      # start a new turn
      linkActor(turn, "http-conn") do (turn: Turn):
        preventInertCheck(turn)
        let facet = turn.facet
        conn.onConnectionError do (err: ref Exception):
          terminateFacet(facet, err)
            # terminate this actor on exception
        # facet is scoped to the lifetime of the connection
        service Session(
            facet: turn.facet,
            driver: driver,
            conn: conn,
            port: port,
          )
  listener

proc httpDriver(turn: Turn; ds: Cap) =
  let driver = Driver(facet: turn.facet, ds: ds, timers: turn.newDataspace)
  spawnTimerDriver(turn, driver.timers)

  during(turn, driver.ds, HttpBinding?:{ 1: grab() }) do (port: BiggestInt):
    publish(turn, ds, HttpListener(port: port))

  # TODO: publish <http-service …>.

  during(turn, driver.ds, ?:HttpListener) do (port: uint16):
    let l = httpListen(turn, driver, Port port)
  do:
    stop(l)

proc spawnHttpDriver*(turn: Turn; ds: Cap): Actor {.discardable.} =
  spawnActor(turn, "http-driver") do (turn: Turn):
    httpDriver(turn, ds)
